module Main where

import qualified Data.ByteString.Builder as BB
import qualified Data.Vector as V
import Diff
import Diff3
import Format
import Hunks
import Merge
import Options.Applicative
import Redfa
import System.IO (stdout)
import System.IO.MMap
import Types
import Version

data ADiffOptions =
  ADiffOptions
    { adiffRedfaOpt :: RedfaOption
    , adiffCmdOpts :: ADiffCommandOpts
    }
  deriving (Show)

data ADiffCommandOpts
  = CmdDiff
      { context :: Int
      , diffFile1 :: String
      , diffFile2 :: String
      }
  | CmdPatch
      { patchDryRun :: Bool
      , patchInDir :: Maybe String
      , patchInput :: String
      , patchReverse :: Bool
      , patchPathStrip :: Int
      , patchMergeOpts :: MergeOpts
      }
  | CmdDiff3
      { context :: Int
      , diff3Mine :: String
      , diff3Old :: String
      , diff3Yours :: String
      , diff3MergeOpts :: MergeOpts
      }
  deriving (Show)

contextOpt :: Parser Int
contextOpt =
  check <$>
  option
    auto
    (metavar "CONTEXT" <>
     short 'C' <>
     long "context" <>
     value 5 <> help "How many tokens around the change to include in the patch")
  where
    check c
      | c < 0 = error "Negative context"
      | otherwise = c

diffCmdOptions :: Parser ADiffCommandOpts
diffCmdOptions =
  CmdDiff <$> contextOpt <*> strArgument (metavar "FROMFILE") <*>
  strArgument (metavar "TOFILE")

patchCmdOptions :: Parser ADiffCommandOpts
patchCmdOptions =
  CmdPatch <$>
  switch
    (short 'n' <>
     long "dry-run" <>
     help "Do not patch anything, just print what would happen") <*>
  optional
    (strOption $
     short 'd' <> long "directory" <> metavar "DIR" <> help "Work in DIR") <*>
  strOption
    (short 'i' <>
     long "input" <>
     metavar "INPUT" <>
     help "Read the patchfile from INPUT, defaults to `-' for STDIN" <>
     value "-") <*>
  switch (short 'R' <> long "reverse" <> help "Unapply applied patches") <*>
  option
    auto
    (short 'p' <>
     long "strip" <>
     metavar "NUM" <>
     help "Strip NUM leading components from the paths" <> value 0) <*>
  mergeOption True

diff3CmdOptions :: Parser ADiffCommandOpts
diff3CmdOptions =
  CmdDiff3 <$> contextOpt <*> strArgument (metavar "MYFILE") <*>
  strArgument (metavar "OLDFILE") <*>
  strArgument (metavar "YOURFILE") <*>
  mergeOption False

actionOption :: Parser ADiffCommandOpts
actionOption =
  hsubparser $
  mconcat
    [ command "diff" $ info diffCmdOptions $ progDesc "Compare two files"
    , command "patch" $ info patchCmdOptions $ progDesc "Apply a patch to files"
    , command "diff3" $
      info diff3CmdOptions $ progDesc "Compare and merge three files"
    ]

adiffOptions :: Parser ADiffOptions
adiffOptions = ADiffOptions <$> redfaOption <*> actionOption

loadToks :: RedfaSpec -> FilePath -> IO TV
loadToks redfa f =
  V.fromList <$> (mmapFileByteString f Nothing >>= redfaTokenize redfa)

main :: IO ()
main =
  let opts :: ParserInfo ADiffOptions
      opts =
        info
          (adiffOptions <**> versionOption "adiff" <**> helperOption)
          (fullDesc <>
           progDesc
             "Compare, patch and merge files on arbitrarily-tokenized sequences." <>
           header "adiff: arbitrary-token diff utilities")
   in do ADiffOptions {adiffRedfaOpt = ropt, adiffCmdOpts = copt} <-
           execParser opts
         redfa <- redfaPrepareRules ropt
         case copt of
           CmdDiff {diffFile1 = f1, diffFile2 = f2, context = ctxt} -> do
             [toks1, toks2] <- traverse (loadToks redfa) [f1, f2]
             BB.hPutBuilder stdout $
               pprHunks $ hunks (max 0 ctxt) $ diffToks toks1 toks2
           CmdPatch {} -> putStrLn "not supported yet"
           CmdDiff3 ctxt f1 f2 f3 mo -> do
             [toksMine, toksOld, toksYour] <-
               traverse (loadToks redfa) [f1, f2, f3]
             let d3 = diff3Toks mo toksMine toksOld toksYour
             BB.hPutBuilder stdout $
               if mergeDoMerge mo
                 then fmtMerged mo d3
                 else pprHunks $ hunks (max 0 ctxt) d3
